#!/usr/bin/env python3

# Sample script for certificate management, using the Avocent(r) ADX REST API
#
# This script is provided "as-is" with no warranty of suitability
# for any particular purpose. The user assumes all risk.
#
# This script may be copied and modified without restriction.

import base64
import re
import requests  # see https://pypi.org/project/requests/
import sys
import time
import urllib3

# Version of this program
VERSION = '1.0.4'

# The name of this program
PROG_NAME = 'adxcert.py'

# Timeout that we use for https requests, in seconds
TIMEOUT_SEC = 60.0

# Recognized command verbs
CMD_VERBS = ['gencsr', 'gencert', 'setcert', 'getdevices']


# Print full help/usage message.
def help(msg=None):
    print('Avocent(r) ADX certificate management tool, v' + VERSION)
    print('Performs various certificate management operations.')
    print('')
    if msg is not None:
        print(msg)
        print('')
    print('Usage:')
    print('  ' + PROG_NAME + ' COMMAND')
    print('')
    print('COMMAND can be:')
    print('  gencsr PARAMS...      Generate and return a Certificate Signing Request.')
    print('  getcert PARAMS...     Get the current certificate.')
    print('  setcert PARAMS...     Install a new certificate.')
    print('  getdevices PARAMS...  Get a list of managed devices.')
    print('')
    print('Get the current certificate and store to the specified file:')
    print('  getcert ip=IPADDRESS user=USERNAME pass=PASSWORD id=DEVICEID file=FILE')
    print('    IPADDRESS        IP address of the management appliance (MP1000 or a standalone RM1048P).')
    print('    USERNAME         username for access to the management appliance.')
    print('    PASSWORD         password for access to the management appliance.')
    print('    DEVICEID         Device ID of the device from which to get the certificate. Use "local" for the management appliance itself.')
    print('    FILE             Name of file to store the retrieved certificate (PEM format). If not supplied, stdout is used.')
    print('')
    print('Generate and get a CSR and store to the specified file:')
    print('  gencsr ip=IPADDRESS user=USERNAME pass=PASSWORD id=DEVICEID subject=SUBJECT san=SAN file=FILE')
    print('    IPADDRESS        IP address of the management appliance (MP1000 or a standalone RM1048P).')
    print('    USERNAME         username for access to the management appliance.')
    print('    PASSWORD         password for access to the management appliance.')
    print('    DEVICEID         Device ID of the device on which to generate the CSR. Use "local" for the management appliance itself.')
    print('    SUBJECT          Subject line for CSR:')
    print('                       /emailAddress=EMAIL/C=COUNTRY/ST=STATE/L=CITY/O=ORG/OU=UNIT/CN=COMNAME')
    print('                         COUNTRY    2-character country code (such as US)')
    print('                         EMAIL      email address (optional)')
    print('                         STATE      state or province (optional)')
    print('                         CITY       city (optional)')
    print('                         ORG        organization name (optional)')
    print('                         UNIT       organization unit name (optional)')
    print('                         COMNAME    common name (such as www.example.com)')
    print('    SAN              Subject Alternative Name line for CSR:')
    print('                       DNS:COMNAME,DNS:ALTNAME,DNS:ALTNAME,IP:ADDR...')
    print('                         COMNAME    alternative name (typically the same as common name (such as www.example.com)')
    print('                         ALTNAME    alternative name (such as www2.example.com)')
    print('                         ADDR       IP address (optional)')
    print('    FILE             Name of file to store the generated CSR (PEM format). If not supplied, stdout is used.')
    print('')
    print('Install the provided certificate (must correspond to the most recently generated CSR):')
    print('  setcert ip=IPADDRESS user=USERNAME pass=PASSWORD id=DEVICEID file=FILE')
    print('    IPADDRESS        IP address of the management appliance (MP1000 or a standalone RM1048P).')
    print('    USERNAME         username for access to the management appliance.')
    print('    PASSWORD         password for access to the management appliance.')
    print('    DEVICEID         Device ID of the device on which to install the certificate. Use "local" for the management appliance itself.')
    print('    FILE             Name of file that contains the certificate (PEM format).')
    print('')
    print('Get the list of devices managed by a management appliance (MP1000 or standalone RM1048P):')
    print('  getdevices ip=IPADDRESS user=USERNAME pass=PASSWORD file=FILE')
    print('    IPADDRESS        IP address of the management appliance (MP1000 or a standalone RM1048P).')
    print('    USERNAME         username for access to the management appliance.')
    print('    PASSWORD         password for access to the management appliance.')
    print('    FILE             Name of file where output should be written. If not supplied, stdout is used.')
    print('')
    print('Examples:')
    print('')
    print('Get the current certificate and store to the specified file:')
    print('  ' + PROG_NAME + ' getcert ip=10.207.15.8 user=admin pass=secret id=local file=10.207.15.8-cert.pem')
    print('')
    print('Generate and get a CSR and store to the specified file:')
    print('  ' + PROG_NAME + ' gencsr ip=10.207.15.8 user=admin pass=secret id=local subject="/emailAddress=me@someplace.com/C=US/ST=Oregon/L=Widget City/O=Advance Widgets/OU=Engineering/CN=www.example.com" san=DNS:www.example.com,IP:10.207.15.8 file=10.207.15.8-csr.pem')
    print('')
    print('Install the provided certificate (must correspond to the previously generated CSR):')
    print('  ' + PROG_NAME + ' setcert ip=10.207.15.8 user=admin pass=secret id=local file=10.207.15.8-newcert.pem')
    print('')
    print('Get the list of devices managed by an appliance:')
    print('  ' + PROG_NAME + ' getdevices ip=10.207.15.8 user=admin pass=secret')
    print('')
    print('Get the current certificate for a managed device, using its device ID:')
    print('  ' + PROG_NAME + ' getcert ip=10.207.15.8 user=admin pass=secret id=563de19b-a5b2-4818-8fc4-015876bdc7f5 file=563de19b-a5b2-4818-8fc4-015876bdc7f5-cert.pem')
    print('')

    sys.exit(1)


# Print brief usage message and exit.
def usage(msg=None):
    print('Performs various certificate management operations.')
    print("Try '" + PROG_NAME + " help' for more information")
    if msg is not None:
        print('\n' + msg)
    sys.exit(1)


# Print message and terminate.
def fatal(msg):
    print('ERROR: ' + msg)
    sys.exit(1)


# Check that all the required parameters are present.
def check_required(params, required):
    for r in required:
        if r not in params:
            usage(f'Missing required parameter "{r}"')


# Check for any unrecognized parameters.
def check_unrecognized(params, recognized):
    for p in params:
        if p not in recognized:
            usage(f'Unrecognized parameter "{p}"')


# Convert an HTTP status code numeric value to interpreted representation.
def pretty_status(code):
    lookup = {
        200: 'OK',
        201: 'OK (created)',
        202: 'OK (accepted)',
        204: 'OK (no content)',
        400: 'Bad Request',
        401: 'Unauthorized',
        403: 'Forbidden',
        404: 'Not Found',
        405: 'Method Not Allowed',
        406: 'Not Acceptable',
        407: 'Proxy Authentication Required',
        408: 'Request Timeout',
        409: 'Conflict',
        410: 'Gone',
        411: 'Length Required',
        413: 'Payload Too Large',
        414: 'URI Too Long',
        415: 'Unsupported Media Type',
        417: 'Expectation Failed',
        426: 'Upgrade Required',
        428: 'Precondition Required',
        429: 'Too Many Requests',
        431: 'Request Header Fields Too Large',
        500: 'Internal Server Error',
        501: 'Not Implemented',
        502: 'Bad Gateway',
        503: 'Service Unavailable',
        504: 'Gateway Timeout',
    }
    if code in lookup:
        out = f'"{lookup[code]}" ({code})'
    else:
        out = str(code)
    return out


# Take a string and return a base64 encoded string.
def base64_encode(input_string):
    input_bytes = input_string.encode("ascii")  # convert to bytes
    base64_bytes = base64.b64encode(input_bytes)  # do the encoding
    output_string = base64_bytes.decode("ascii")  # convert to string
    return output_string


# Take a base64-encoded string and return a decoded string.
def base64_decode(input_string):
    input_bytes = input_string.encode("ascii")  # convert to bytes
    decoded_bytes = base64.b64decode(input_bytes)  # decode
    output_string = decoded_bytes.decode("ascii")  # convert to string
    return output_string


# Authenticate and return the JWT.
def authenticate(ip, user, password):
    url = f'https://{ip}/api/v1/userSessions'
    try:
        response = requests.post(url, json={'username': user, 'password': password}, timeout=TIMEOUT_SEC, verify=False)
    except (requests.exceptions.ConnectTimeout, requests.exceptions.SSLError):
        fatal(f'Communication error with "{ip}" during authentication. Perhaps server is down or restarting.')
    if response.status_code == 201:
        json_response = response.json()
        if 'jwt' not in json_response:
            fatal('Authentication response did not contain jwt.')
        jwt = json_response['jwt']
        return jwt
    return None


# Convert the subject string to dictionary form. Example:
# '/emailAddress=me@example.com/C=US/ST=Idaho/L=Boise/O=MWC/OU=My Dept/CN=www.example.com'
def subject_to_dict(subject):
    dict = {}

    r = re.search(r'.*/emailAddress=([^/]+)', subject)
    if r:
        dict["email_address"] = r.group(1)

    r = re.search(r'.*/C=([^/]+)', subject)
    if r:
        dict["country_name"] = r.group(1)

    r = re.search(r'.*/ST=([^/]+)', subject)
    if r:
        dict["state_or_province_name"] = r.group(1)

    r = re.search(r'.*/L=([^/]+)', subject)
    if r:
        dict["locality_name"] = r.group(1)

    r = re.search(r'.*/O=([^/]+)', subject)
    if r:
        dict["organization_name"] = r.group(1)

    r = re.search(r'.*/OU=([^/]+)', subject)
    if r:
        dict["organizational_unit_name"] = r.group(1)

    r = re.search(r'.*/CN=([^/]+)', subject)
    if r:
        dict["common_name"] = r.group(1)

    return dict


# Generate a CSR.
def gen_csr(params):
    check_required(params, ['ip', 'user', 'pass', 'id', 'subject'])
    check_unrecognized(params, ['ip', 'user', 'pass', 'id', 'subject', 'san', 'file'])
    ip = params['ip']
    user = params['user']
    pw = params['pass']
    id = params['id']
    subject = params['subject']
    params_dict = subject_to_dict(subject)
    if 'san' in params:
        san = params['san']
        params_dict['subject_alternative_names'] = san
    else:
        san = None
    if 'file' in params:
        filename = params['file']
    else:
        filename = None
    jwt = authenticate(ip, user, pw)
    if jwt is None:
        fatal("Authentication failed.")
    headers = {"Authorization": "Bearer " + jwt}
    url = f"https://{ip}/api/v1/devices/{id}:generateCsr"
    try:
        response = requests.post(url, json=params_dict, timeout=TIMEOUT_SEC, headers=headers, verify=False)
    except (requests.exceptions.ConnectTimeout, requests.exceptions.ReadTimeout):
        fatal(f'Communication error with "{ip}" during POST. Perhaps server is down or restarting.')
    if response.status_code == 404:
        fatal(f'Server responded with {pretty_status(response.status_code)} when generating CSR. Check ID "{id}".')
    if response.status_code != 200:
        fatal(f'Problem encountered when generating CSR. Server responded with {pretty_status(response.status_code)}.')
    # Now we need to poll for completion.
    for tries in range(1, 10):
        url = f"https://{ip}/api/v1/devices/{id}/csr"
        try:
            response = requests.get(url, timeout=TIMEOUT_SEC, headers=headers, verify=False)
        except (requests.exceptions.ConnectTimeout, requests.exceptions.ReadTimeout):
            fatal(f'Connection to "{ip}" timed out during GET while checking for CSR completion.')
        json_response = response.json()
        try:
            status = json_response['status']
            # If it is done, then get the data and write to the desired file.
            if status == 'CSR_GENERATION_STATUS_COMPLETED':
                csr = json_response['csr']
                csr_data = csr['encoded_data']  # this is base64 encoded
                decoded_csr_data = base64_decode(csr_data)
                if filename is not None:
                    try:
                        file = open(filename, 'w')
                        file.write(decoded_csr_data)
                        file.close()
                    except IOError:
                        fatal(f'Unable to write CSR to output file "{filename}".')
                else:
                    print(decoded_csr_data)
                return  # all good, if we get here
            else:
                time.sleep(2)  # delay slightly, and try again
                continue
        except KeyError as e:
            fatal(f'Response does not contain expected key ({str(e)}) when getting CSR.')


# Get a certificate.
def get_cert(params):
    check_required(params, ['ip', 'user', 'pass', 'id'])
    check_unrecognized(params, ['ip', 'user', 'pass', 'id', 'file'])
    ip = params['ip']
    user = params['user']
    pw = params['pass']
    id = params['id']
    if 'file' in params:
        filename = params['file']
    else:
        filename = None
    jwt = authenticate(ip, user, pw)
    if jwt is None:
        fatal("Authentication failed.")
    headers = {"Authorization": "Bearer " + jwt}
    url = f"https://{ip}/api/v1/devices/{id}/config?filter=certificate"
    try:
        response = requests.get(url, timeout=TIMEOUT_SEC, headers=headers, verify=False)
    except (requests.exceptions.ConnectTimeout, requests.exceptions.ReadTimeout):
        fatal(f'Communication error with "{ip}" during GET. Perhaps server is down or restarting.')
    if response.status_code == 200:
        json_response = response.json()
        found = False
        try:
            settings = json_response['settings']
            for setting in settings:
                if 'certificate' in setting:
                    certificate = setting['certificate']
                    if 'encoded_data' in certificate:
                        encoded_data = certificate['encoded_data']
                        found = True
                        break
        except KeyError as e:
            fatal(f'Response does not contain expected key ({str(e)}) when getting certificate.')
        if not found:
            fatal("Response does not contain expected key ('certificate') when getting certificate.")
        certificate_data = base64_decode(encoded_data)
        if '-----BEGIN CERTIFICATE-----' not in certificate_data:  # sanity check
            fatal('Server did not return valid certificate data.')
        if filename is not None:
            try:
                file = open(filename, 'w')
                file.write(certificate_data)
                file.close()
            except IOError:
                fatal(f'Unable to write certificate to output file "{filename}".')
        else:
            print(certificate_data)
        return  # all good, if we get here
    elif response.status_code == 404:
        fatal(f'Server responded with {pretty_status(response.status_code)} when getting certificate. Check ID "{id}".')
    fatal(f'Problem encountered when getting certificate. Server responded with {pretty_status(response.status_code)}.')


# Set (update) the certificate.
def set_cert(params):
    check_required(params, ['ip', 'user', 'pass', 'id', 'file'])
    check_unrecognized(params, ['ip', 'user', 'pass', 'id', 'file'])
    ip = params['ip']
    user = params['user']
    pw = params['pass']
    id = params['id']
    filename = params['file']

    # Read certificate from file and prepare for sending.
    try:
        file = open(filename, 'r')
        certificate_data = file.read()
        file.close()
    except Exception as e:
        fatal(f'Failed to read certificate file "{filename}". Exception: {str(e)}')
    base64_data = base64_encode(certificate_data)
    json_payload = {'certificate': {'encoded_data': base64_data}}

    jwt = authenticate(ip, user, pw)
    if jwt is None:
        fatal("Authentication failed.")
    headers = {"Authorization": "Bearer " + jwt}
    url = f"https://{ip}/api/v1/devices/{id}/config"
    try:
        response = requests.patch(url, json=json_payload, timeout=TIMEOUT_SEC, headers=headers, verify=False)
    except (requests.exceptions.ConnectTimeout, requests.exceptions.ReadTimeout, requests.exceptions.ConnectionError):
        fatal(f'Communication error with "{ip}" during PATCH. Perhaps server is down or restarting.')
    if response.status_code == 404:
        fatal(f'Server responded with {pretty_status(response.status_code)} when updating certificate. Check ID "{id}".')
    if response.status_code != 200:
        # Get message from response body, if there is one.
        err_message = f'Problem encountered updating certificate. Server responded with {pretty_status(response.status_code)}.'
        json_response = response.json()
        if 'message' in json_response:
            message = json_response['message']
            err_message += ' "' + message + '"'
        fatal(err_message)


# Get a list of managed devices
def get_devices(params):
    check_required(params, ['ip', 'user', 'pass'])
    check_unrecognized(params, ['ip', 'user', 'pass', 'file'])
    ip = params['ip']
    user = params['user']
    pw = params['pass']
    if 'file' in params:
        filename = params['file']
    else:
        filename = None
    jwt = authenticate(ip, user, pw)
    if jwt is None:
        fatal("Authentication failed.")
    headers = {"Authorization": "Bearer " + jwt}
    url = f"https://{ip}/api/v1/devices"
    try:
        response = requests.get(url, timeout=TIMEOUT_SEC, headers=headers, verify=False)
    except (requests.exceptions.ConnectTimeout, requests.exceptions.ReadTimeout):
        fatal(f'Communication error with "{ip}" during GET. Perhaps server is down or restarting.')
    if response.status_code != 200:
        fatal(f'Problem encountered when getting device list. Server responded with {pretty_status(response.status_code)}.')
    json_response = response.json()
    if 'devices' not in json_response:
        fatal("Response does not contain expected key ('devices') when getting device list.")
    devices = json_response['devices']

    # Convert the returned data into format suitable for our output.
    try:
        out = ''
        for device in devices:
            id = device['id']
            device_name = device['device_name']
            device_type_name = device['device_type_name']
            device_status = device['device_status']
            out += f'"id"="{id}",'
            out += f'"device_name"="{device_name}",'
            out += f'"device_type_name"="{device_type_name}",'
            out += f'"device_status"="{device_status}"\n'
    except KeyError as e:
        fatal(f'Unable to decode response. Missing key {str(e)}.')

    # Write to the output file.
    if filename is not None:
        try:
            file = open(filename, 'w')
            file.write(out)
            file.close()
        except IOError:
            fatal(f'Unable to write device list to output file "{filename}".')
    else:
        print(out)


# Convert a list of parameters into a corresponding dictionary, lowercasing the key. ['A=b','c=d'] => {'a':'b','c':'d'}
def dict_of_params(arguments):
    RE_ARG = re.compile(r'([^=]+)=(.+)')
    d = {}
    for arg in arguments:
        s = RE_ARG.search(arg)
        if s is not None:
            left = s.group(1).lower()
            right = s.group(2)
            d[left] = right
    return d


# Main
def main():

    # Disable certificate verification warnings, because the devices use self-signed certificates by default. Without
    # disabling this, we would repeatedly see a message as this script is being used:
    #   "InsecureRequestWarning: Unverified HTTPS request is being made to host"
    # For more information, see https://urllib3.readthedocs.io/en/latest/advanced-usage.html#ssl-warnings
    # If you want to see the warnings, just remove or comment out the line below.
    urllib3.disable_warnings()

    # If no parameters supplied, show usage message.
    if len(sys.argv) == 1:
        usage()

    # Get the command verb, converting to lowercase so we can accept either upper or lowercase from the user.
    cmd = sys.argv[1].lower()

    # Is user asking for command help?
    if cmd == 'help' or cmd == '--help' or cmd == '-help' or cmd == '-h' or cmd[0] == '?':
        help()

    # Make a dictionary of the parameters.
    params = dict_of_params(sys.argv[2:])

    # Call the appropriate function for the desired command.
    if cmd == 'gencsr':
        gen_csr(params)
    elif cmd == 'getcert':
        get_cert(params)
    elif cmd == 'setcert':
        set_cert(params)
    elif cmd == 'getdevices':
        get_devices(params)
    else:
        usage('Unrecognized command')


# Run main()
if __name__ == '__main__':
    main()
